use evdev::{Device as EvDevice, EventType, InputEventKind};
use std::collections::HashMap;

use crate::graph_types::{Descriptor, NodeDescriptor};

pub struct Device {
    device: EvDevice,
    path: String,

    supported_events: HashMap<EventType, Vec<InputEventKind>>,
}

impl Device {
    pub fn query() -> Vec<Device> {
        evdev::enumerate()
            .map(|(path, device)| {
                let supported_events = device
                    .supported_events()
                    .iter()
                    .map(|event| {
                        let input_kind: Vec<InputEventKind> = match event {
                            evdev::EventType::KEY => device
                                .supported_keys()
                                .map(|keys| {
                                    keys.iter().map(|key| InputEventKind::Key(key)).collect()
                                })
                                .unwrap_or(Vec::new()),
                            evdev::EventType::RELATIVE => device
                                .supported_relative_axes()
                                .map(|axes| {
                                    axes.iter()
                                        .map(|rel_axis| InputEventKind::RelAxis(rel_axis))
                                        .collect()
                                })
                                .unwrap_or(Vec::new()),
                            evdev::EventType::ABSOLUTE => device
                                .supported_absolute_axes()
                                .map(|axes| {
                                    axes.iter()
                                        .map(|abs_axis| InputEventKind::AbsAxis(abs_axis))
                                        .collect()
                                })
                                .unwrap_or(Vec::new()),
                            evdev::EventType::SWITCH => device
                                .supported_switches()
                                .map(|switches| {
                                    switches
                                        .iter()
                                        .map(|switch| InputEventKind::Switch(switch))
                                        .collect()
                                })
                                .unwrap_or(Vec::new()),
                            evdev::EventType::MISC => device
                                .misc_properties()
                                .map(|miscs| {
                                    miscs
                                        .iter()
                                        .map(|misc| InputEventKind::Misc(misc))
                                        .collect()
                                })
                                .unwrap_or(Vec::new()),
                            evdev::EventType::SOUND => device
                                .supported_sounds()
                                .map(|sounds| {
                                    sounds
                                        .iter()
                                        .map(|sound| InputEventKind::Sound(sound))
                                        .collect()
                                })
                                .unwrap_or(Vec::new()),
                            evdev::EventType::FORCEFEEDBACK => device
                                .supported_forcefeedback_effects()
                                .map(|ffs| {
                                    ffs.iter()
                                        .map(|ff| InputEventKind::ForceFeedback(ff))
                                        .collect()
                                })
                                .unwrap_or(Vec::new()),

                            _ => Vec::new(),
                        };

                        (event, input_kind)
                    })
                    .collect();

                Device {
                    device,
                    path: path.as_path().to_str().unwrap().to_string(),

                    supported_events,
                }
            })
            .collect()
    }

    pub fn name(&self) -> &str {
        self.device.name().unwrap_or("Unnamed Device")
    }

    pub fn to_descriptor(&self) -> NodeDescriptor {
        let mut inputs = match self.supported_events.get(&EventType::FORCEFEEDBACK) {
            Some(ff) => ff
                .iter()
                .map(|kind| {
                    (
                        format!(
                            "{:?}",
                            match kind {
                                InputEventKind::ForceFeedback(ff) => ff,
                                _ => panic!("wrong kind"),
                            }
                        ),
                        // TODO: find min and max values
                        Descriptor::Slider { min: 0, max: 255 },
                    )
                })
                .collect(),
            None => Vec::new(),
        };

        let outputs = Self::flatten(
            self.supported_events
                .iter()
                .map(|(event_type, kinds)| {
                    let t: Option<Vec<(String, Descriptor)>> = match *event_type {
                        EventType::ABSOLUTE => Some(
                            kinds
                                .iter()
                                .map(|kind| {
                                    (
                                        format!(
                                            "{:?}",
                                            match kind {
                                                InputEventKind::AbsAxis(axis) => axis,
                                                _ => panic!("wrong kind"),
                                            }
                                        ),
                                        // TODO: find min and max values
                                        Descriptor::Slider { min: 0, max: 255 },
                                    )
                                })
                                .collect(),
                        ),
                        EventType::RELATIVE => Some(
                            kinds
                                .iter()
                                .map(|kind| {
                                    (
                                        format!(
                                            "{:?}",
                                            match kind {
                                                InputEventKind::RelAxis(axis) => axis,
                                                _ => panic!("wrong kind"),
                                            }
                                        ),
                                        // TODO: find min and max values
                                        Descriptor::Slider { min: 0, max: 255 },
                                    )
                                })
                                .collect(),
                        ),
                        EventType::KEY => Some(
                            kinds
                                .iter()
                                .map(|kind| {
                                    (
                                        format!(
                                            "{:?}",
                                            match kind {
                                                InputEventKind::Key(key) => key,
                                                _ => panic!("wrong kind"),
                                            }
                                        ),
                                        Descriptor::Button,
                                    )
                                })
                                .collect(),
                        ),

                        _ => None,
                    };

                    t
                })
                .filter_map(|f| f)
                .collect::<Vec<Vec<(String, Descriptor)>>>(),
        );

        inputs.insert(0, (format!("Path: {}", self.path), Descriptor::Info));

        NodeDescriptor {
            op_name: self.name().to_string(),
            inputs,
            outputs,
        }
    }

    fn flatten<T>(vv: Vec<Vec<T>>) -> Vec<T> {
        vv.into_iter().flatten().collect()
    }
}

impl std::fmt::Debug for Device {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("Device")
            .field("name", &self.device.name())
            .field("path", &self.path)
            .field("supported_events", &self.supported_events)
            .finish()
    }
}
