use egui::*;

use crate::color_hex_utils::*;
use crate::device::Device;
use crate::graph_types::NodeDescriptor;
use crate::virtual_device::VirtualDevice;

pub enum SelectedDevice<'a> {
    Virtual(VirtualDevice),
    Physical(&'a Device),
}

impl<'a> SelectedDevice<'a> {
    pub fn to_descriptor(&self) -> NodeDescriptor {
        match self {
            SelectedDevice::Virtual(dev) => dev.to_descriptor(),
            SelectedDevice::Physical(dev) => dev.to_descriptor(),
        }
    }
}

#[derive(Default)]
pub struct NodeFinder {
    query: String,
    /// Reset every frame. When set, the node finder will be moved at that position
    pub position: Option<Pos2>,
    pub just_spawned: bool,
}

impl NodeFinder {
    pub fn new_at(pos: Pos2) -> Self {
        NodeFinder {
            position: Some(pos),
            just_spawned: true,
            ..Default::default()
        }
    }

    /// Shows the node selector panel with a search bar. Returns whether a node
    /// archetype was selected and, in that case, the finder should be hidden on
    /// the next frame.
    pub fn show<'a>(
        &mut self,
        devices: &'a [Device],
        allow_virt_dev: bool,
        ui: &mut Ui,
    ) -> Option<SelectedDevice<'a>> {
        let background_color = color_from_hex("#3f3f3f").unwrap();
        let _titlebar_color = background_color.linear_multiply(0.8);
        let text_color = color_from_hex("#fefefe").unwrap();

        ui.visuals_mut().widgets.noninteractive.fg_stroke = Stroke::new(2.0, text_color);

        let frame = Frame::dark_canvas(ui.style())
            .fill(background_color)
            .margin(vec2(5.0, 5.0));

        // The archetype that will be returned.
        let mut submitted_device = None;
        frame.show(ui, |ui| {
            ui.vertical(|ui| {
                Frame::default().margin(vec2(10.0, 10.0)).show(ui, |ui| {
                    if allow_virt_dev {
                        if ui.selectable_label(false, "New Virtual Device").clicked() {
                            submitted_device = Some(SelectedDevice::Virtual(VirtualDevice::new(
                                "New Virtual Device",
                            )));
                        }
                    }

                    for device in devices {
                        let device_name = device.name();
                        if device_name.contains(self.query.as_str()) {
                            if ui.selectable_label(false, device_name).clicked() {
                                submitted_device = Some(SelectedDevice::Physical(device));
                            }
                        }
                    }
                });
            });
        });

        submitted_device
    }
}
